@@ -1,17 +1,17 @@
 
   (**************************************************************************)
   (*  ocaml-dtools                                                          *)
-  (*  Copyright (C) 2003-2006  The Savonet Team                             *)
+  (*  Copyright (C) 2003-2010  The Savonet Team                             *)
   (**************************************************************************)
   (*  This program is free software; you can redistribute it and/or modify  *)
   (*  it under the terms of the GNU General Public License as published by  *)
   (*  the Free Software Foundation; either version 2 of the License, or     *)
   (*  any later version.                                                    *)
   (**************************************************************************)
-  (*  Contact: dev@gim.name                                                 *)
+  (*  Contact: savonet-devl@lists.sourceforge.net                           *)
   (**************************************************************************)
 
-(* $Id: dtools.ml 5102 2008-02-09 14:19:41Z metamorph68 $ *)
+(* $Id$ *)
 
 (**
   ocaml-dtools
@@ -382,8 +382,12 @@ struct
 
   let conf =
     Conf.void "initialization configuration"
+  (* Unix.fork is not implemented in Win32. *)
+  let daemon_conf = 
+    if Sys.os_type <> "Win32" then conf
+    else Conf.void "dummy conf"
   let conf_daemon =
-    Conf.bool ~p:(conf#plug "daemon") ~d:false
+    Conf.bool ~p:(daemon_conf#plug "daemon") ~d:false
       "run in daemon mode"
   let conf_daemon_pidfile =
     Conf.bool ~p:(conf_daemon#plug "pidfile") ~d:false
@@ -391,6 +395,15 @@ struct
   let conf_daemon_pidfile_path =
     Conf.string ~p:(conf_daemon_pidfile#plug "path")
       "path to pidfile"
+  let conf_daemon_drop_user = 
+    Conf.bool ~p:(conf_daemon#plug "change_user") ~d:false
+      "Changes the effective user (drops privileges)."
+  let conf_daemon_user = 
+    Conf.string ~p:(conf_daemon_drop_user#plug "user") 
+      ~d:"daemon" "User used to run the daemon."
+  let conf_daemon_group = 
+    Conf.string ~p:(conf_daemon_drop_user#plug "group") 
+      ~d:"daemon" "Group used to run the daemon."
   let conf_trace =
     Conf.bool ~p:(conf#plug "trace") ~d:false
       "dump an initialization trace"
@@ -499,21 +512,36 @@ struct
   exception StartError of exn
   exception StopError of exn
 
+  (* Dummy functions in the case where 
+   * Printexc does not have the required
+   * functions. *)
+  let get_backtrace () =
+  "ocaml-dtools not compiled with ocaml >= 3.11, \
+   cannot print stack backtrace"
+
+  open Printexc
+
   let main f () =
     begin try exec start with e -> raise (StartError e) end;
-    let quit pid = Unix.kill pid Sys.sigterm in
+    let quit pid = 
+      if Sys.os_type <> "Win32" then
+        Unix.kill pid Sys.sigterm 
+    in
     let thread pid =
       begin try f (); quit pid with
       | e ->
 	  let se = Printexc.to_string e in
 	  Printf.eprintf
 	    "init: exception encountered during main phase:\n  %s\n%!" se;
-	  Printf.eprintf "exception: %s\n%!" se;
+	  Printf.eprintf "exception: %s\n%s%!" se (get_backtrace ());
 	  if conf_catch_exn#get then quit pid else raise e
       end
     in
-    ignore (Thread.create thread (Unix.getpid ()));
-    wait_signal ();
+    let th = Thread.create thread (Unix.getpid ()) in
+    if Sys.os_type <> "Win32" then
+      wait_signal ()
+    else
+      Thread.join th ;
     begin try exec stop with e -> raise (StopError e) end
 
   let catch f clean =
@@ -535,14 +563,16 @@ struct
     end
 
   (** A function to reopen a file descriptor
-    * Thanks to Xavier Leroy !
-    * Ref: http://caml.inria.fr/pub/ml-archives/caml-list/2000/01/a7e3bbdfaab33603320d75dbdcd40c37.en.html
+    * Thanks to Xavier Leroy!
+    * Ref: http://caml.inria.fr/pub/ml-archives/caml-list/2000/01/
+    *      a7e3bbdfaab33603320d75dbdcd40c37.en.html
     *)
   let reopen_out outchan filename =
     flush outchan;
     let fd1 = Unix.descr_of_out_channel outchan in
     let fd2 =
-      Unix.openfile filename [Unix.O_WRONLY; Unix.O_CREAT; Unix.O_TRUNC] 0o666 in
+      Unix.openfile filename [Unix.O_WRONLY] 0o666
+    in
     Unix.dup2 fd2 fd1;
     Unix.close fd2
 
@@ -550,46 +580,50 @@ struct
   let reopen_in inchan filename =
     let fd1 = Unix.descr_of_in_channel inchan in
     let fd2 =
-      Unix.openfile filename [Unix.O_WRONLY; Unix.O_CREAT; Unix.O_TRUNC] 0o666 in
+      Unix.openfile filename [Unix.O_RDONLY] 0o666
+    in
     Unix.dup2 fd2 fd1;
     Unix.close fd2
 
-  let daemonize fn =
-    begin match Unix.fork () with
-    | 0 ->
-        (* Change umask to 0 *)
-        ignore(Unix.umask 0) ;
-	let f =  
-          match conf_daemon_pidfile#get with
-	    | false ->
-	       fun () -> catch fn (fun () -> ())
-	    | true ->
-	       (* Write PID to file *)
-	       let filename = conf_daemon_pidfile_path#get in
-	       let f = open_out filename in
-	       let pid = Unix.getpid () in
-	       output_string f (string_of_int pid);
-	       output_char f '\n';
-	       close_out f ;
-               fun () -> catch fn (fun () -> Unix.unlink filename)
-	in
-        (* Dettach from the console *)
-        if (Unix.setsid () < 0) then
-               exit 1;
-        (* chdir to / *)
-        Unix.chdir "/" ;
-        (* Reopen usual file descriptor *)
-        flush_all ();
-        reopen_in stdin "/dev/null";
-        reopen_out stdout "/dev/null";
-        reopen_out stderr "/dev/null";
-        (* Main loop *)
-        f () ;
-        exit 0
-    | _ -> exit 0
-    end
+  let daemonize () =
+    if Unix.fork () <> 0 then exit 0 ;
+    (* Dettach from the console *)
+    if (Unix.setsid () < 0) then
+           exit 1;
+    (* Refork.. *)
+    if Unix.fork () <> 0 then exit 0 ;
+    (* Change umask to 0 *)
+    ignore(Unix.umask 0) ;
+    (* chdir to / *)
+    Unix.chdir "/" ;
+    if conf_daemon_pidfile#get then
+     begin
+      (* Write PID to file *)
+      let filename = conf_daemon_pidfile_path#get in
+      let f = open_out filename in
+      let pid = Unix.getpid () in
+      output_string f (string_of_int pid);
+      output_char f '\n';
+      close_out f 
+     end ;
+    (* Reopen usual file descriptor *)
+    reopen_in stdin "/dev/null";
+    reopen_out stdout "/dev/null";
+    reopen_out stderr "/dev/null"
 
   let exit_when_root () =
+    (* Change user.. *)
+    if conf_daemon_drop_user#get then
+     begin
+      let grd = Unix.getgrnam conf_daemon_group#get in
+      let gid = grd.Unix.gr_gid in
+      if Unix.getegid () <> gid then
+        Unix.setgid gid ;
+      let pwd = Unix.getpwnam conf_daemon_user#get in
+      let uid = pwd.Unix.pw_uid in
+      if Unix.geteuid () <> uid then
+        Unix.setuid uid
+     end;
     let security s = Printf.eprintf "init: security exit, %s\n%!" s in
     if Unix.geteuid () = 0 then
       begin security "root euid (user)."; exit (-1) end;
@@ -598,19 +632,27 @@ struct
 
   let init ?(prohibit_root=false) f =
     if prohibit_root then exit_when_root ();
+    if conf_daemon#get && Sys.os_type <> "Win32" then
+      daemonize () ;
     let signal_h i = () in
     Sys.set_signal Sys.sigterm (Sys.Signal_handle signal_h);
     Sys.set_signal Sys.sigint (Sys.Signal_handle signal_h);
-    if conf_daemon#get
-    then daemonize (main f)
-    else catch (main f) (fun () -> ())
+    (* We block signals that would kill us,
+     * we'll wait for them and shutdown cleanly.
+     * On Windows this is impossible so the only way for the application
+     * to shutdown is to terminate the main function [f]. *)
+    if Sys.os_type <> "Win32" then
+      ignore (Unix.sigprocmask Unix.SIG_BLOCK [Sys.sigterm; Sys.sigint]);
+    catch (main f) (fun () -> ())
 
   let args =
-    [
-      ["--daemon";"-d"],
-      Arg.Unit (fun () -> conf_daemon#set true),
-      "run in daemon mode";
-    ]
+    if Sys.os_type <> "Win32" then
+      [
+        ["--daemon";"-d"],
+        Arg.Unit (fun () -> conf_daemon#set true),
+        "run in daemon mode";
+      ]
+    else []
 
 end
 
@@ -623,11 +665,23 @@ struct
 	f: 'a. int -> ('a, unit, string, unit) format4 -> 'a;
       >
 
+  type custom_log = 
+    { 
+      timestamp : bool ;
+      exec      : string -> unit 
+    }
+
   let log_ch = ref None
 
   (* Mutex to avoid interlacing logs *)
   let log_mutex = Mutex.create ()
 
+  (* Custom logging methods. *)
+  let custom_log : (string, custom_log) Hashtbl.t = Hashtbl.create 0
+
+  let add_custom_log name f = Hashtbl.replace custom_log name f
+  let rm_custom_log name = Hashtbl.remove custom_log name
+
   let conf =
     Conf.void "log configuration"
 
@@ -680,22 +734,27 @@ struct
   let print (time, str) =
     let to_stdout = conf_stdout#get in
     let to_file = !log_ch <> None in
+    let timestamp = timestamp time in
+    let message = 
+        Printf.sprintf "%s %s" timestamp str
+    in
     begin match to_stdout || to_file with
     | true ->
-        let timestamp = timestamp time in
 	let do_stdout () =
-	  Printf.printf "%s %s\n%!" timestamp str;
+	  Printf.printf "%s\n%!" message;
 	in
 	let do_file () =
 	  begin match !log_ch with
 	  | None -> ()
-	  | Some ch -> Printf.fprintf ch "%s %s\n%!" timestamp str;
+	  | Some ch -> Printf.fprintf ch "%s\n%!" message;
 	  end
 	in
 	if to_stdout then do_stdout ();
 	if to_file then do_file ();
     | false -> ()
-    end
+    end ;
+    let f _ x = x.exec (if x.timestamp then message else str) in
+    Hashtbl.iter f custom_log
 
   let proceed entry =
     mutexify (fun () ->
@@ -768,7 +827,8 @@ struct
 	let log_file_path = conf_file_path#get in
 	let log_file_perms = conf_file_perms#get in
 	(* Re-open log file on SIGUSR1 -- for logrotate *)
-	Sys.set_signal Sys.sigusr1
+        if Sys.os_type <> "Win32" then
+	 Sys.set_signal Sys.sigusr1
 	  (Sys.Signal_handle
               begin fun _ ->
 		begin match !log_ch with
