@@ -1,17 +1,17 @@
 
   (**************************************************************************)
   (*  ocaml-dtools                                                          *)
-  (*  Copyright (C) 2003-2006  The Savonet Team                             *)
+  (*  Copyright (C) 2003-2010  The Savonet Team                             *)
   (**************************************************************************)
   (*  This program is free software; you can redistribute it and/or modify  *)
   (*  it under the terms of the GNU General Public License as published by  *)
   (*  the Free Software Foundation; either version 2 of the License, or     *)
   (*  any later version.                                                    *)
   (**************************************************************************)
-  (*  Contact: dev@gim.name                                                 *)
+  (*  Contact: savonet-devl@lists.sourceforge.net                           *)
   (**************************************************************************)
 
-(* $Id: dtools.mli 4647 2007-10-06 16:35:48Z dbaelde $ *)
+(* $Id$ *)
 
 (**
   ocaml-dtools.
@@ -255,6 +255,22 @@ sig
        Type for loggers.
     *)
 
+  type custom_log =
+    {
+      timestamp : bool ;
+      exec      : string -> unit
+    }
+
+  val add_custom_log : string -> custom_log -> unit
+    (**
+      Add a custom logging functions. 
+    *)
+
+  val rm_custom_log : string -> unit
+    (**
+      Remove a custom logging functions.    
+    *)
+
   val make : Conf.path -> t
     (**
       Make a logger labeled according to the given path.
